from .client import Client, unregister_all, register_clients
from .utils import (
    get_all_wiz_handles,
    count_wiz_clients,
    finish_all_loading,
    XYZYaw,
    run_threads,
)
from .card import Card
from .battle import Battle
from .mouse import Mouse
from .window import Window
from .keyboard import Keyboard
from .pixels import DeviceContext, match_image
from .hotkey import HotkeyEvents

# Clean up on exit
import ctypes
import ctypes.wintypes
import asyncio


def close_handler(dwCtrlType):
    asyncio.run(unregister_all())
    return False


handler_func_type = ctypes.WINFUNCTYPE(ctypes.wintypes.BOOL, ctypes.wintypes.DWORD)

transformed_callback = handler_func_type(close_handler)

# https://docs.microsoft.com/en-us/windows/console/setconsolectrlhandler
ctypes.windll.kernel32.SetConsoleCtrlHandler(transformed_callback, True)



import ctypes
from ctypes import WinDLL

user32 = ctypes.WinDLL("user32.dll")


def screen_size():
    """Returns the width and height of the screen as a two-integer tuple.
    Returns:
      (width, height) tuple of the screen size, in pixels.
    """
    return (
        ctypes.windll.user32.GetSystemMetrics(0),
        ctypes.windll.user32.GetSystemMetrics(1),
    )


class Window:
    """
    Base class for all classes in wizSDK. Keeps track of the wizard101 app window.
    """

    def __init__(self, handle=None):
        # If window_handle is None, Window represents the screen
        self.window_handle = handle

    def is_active(self) -> bool:
        """ Returns true if the window is focused """
        if self.window_handle:
            return self.window_handle == user32.GetForegroundWindow()
        else:
            # The "screen" is always active
            return True

    def set_active(self):
        """ Sets the window to active if it isn't already """
        if self.window_handle and not self.is_active():
            user32.SetForegroundWindow(self.window_handle)
        return self

    def get_rect(self) -> tuple:
        """
        Gets the area rectangle of the window (x, y, width, height) relative to the monitor position.
        
        Returns: 
            tuple (x, y, width, height) of the window
        """
        if self.window_handle:
            rect = ctypes.wintypes.RECT()
            user32.GetWindowRect(self.window_handle, ctypes.byref(rect))
            # Returns (x, y, w, h) tuple
            return (rect.left, rect.top, rect.right - rect.left, rect.bottom - rect.top)
        else:
            # Return rect of screen
            return (0, 0, *screen_size())



# Native imports
import ctypes, os
from collections import namedtuple
import asyncio

user32 = ctypes.windll.user32

XYZYaw = namedtuple("XYZYaw", "x y z yaw")
"""
Used to store player location

:meta private:
"""


def run_threads(*coroutines, return_when=asyncio.FIRST_COMPLETED):
    """
    creates an asyncio event loop to run coroutines concurrently
    
    Args:
        coroutines: any amount of coroutines to run at the same time
        return_when: when to stop the threads and return: asyncio.FIRST_COMPLETED, asyncio.ALL_COMPLETED, or asyncio.FIRST_EXCEPTION

    """

    loop = asyncio.get_event_loop()
    job = asyncio.wait(coroutines, return_when=return_when)

    done, pending = loop.run_until_complete(job)

    # finish / cancel all tasks properly
    # https://stackoverflow.com/a/62443715/10751635
    for t in pending:
        t.cancel()

    while not all([t.done() for t in pending]):
        loop._run_once()


def get_all_wiz_handles() -> list:
    """
    Retrieves all window handles for windows that have the 
    'Wizard Graphical Client' class
    
    Returns:
        List of all the wizard101 device handles
    """
    target_class = "Wizard Graphical Client"

    handles = []

    # callback takes a window handle and an lparam and returns true/false on if we should keep going
    # iterating
    # https://docs.microsoft.com/en-us/previous-versions/windows/desktop/legacy/ms633498(v=vs.85)
    def callback(handle, _):
        class_name = ctypes.create_unicode_buffer(len(target_class))
        # win_title = ctypes.create_unicode_buffer(100)

        user32.GetClassNameW(handle, class_name, len(target_class) + 1)
        # user32.GetWindowTextW(handle, win_title, 101)

        if target_class == class_name.value:
            handles.append(handle)

        # iterate all windows
        return 1

    # https://docs.python.org/3/library/ctypes.html#callback-functions
    enumwindows_func_type = ctypes.WINFUNCTYPE(
        ctypes.c_bool,  # return type
        ctypes.c_int,  # arg1 type
        ctypes.POINTER(ctypes.c_int),  # arg2 type
    )

    # Transform callback into a form we can pass to the dll
    callback = enumwindows_func_type(callback)

    # EnumWindows takes a callback every iteration is passed to
    # and an lparam
    # https://docs.microsoft.com/en-us/windows/win32/api/winuser/nf-winuser-enumwindows
    user32.EnumWindows(callback, 0)

    return handles


def count_wiz_clients() -> int:
    """
    Returns the number of wizard101 clients detected
    
    Returns:
        Number of wizard101 clients detected
    """
    return len(get_all_wiz_handles())


async def finish_all_loading(*players):
    """
    Wait for all players passed in as arguments to have gone through the loading screen.
    """
    await asyncio.gather(*[player.finish_loading() for player in players])


def packaged_img(filename: str = ""):
    """
    Helper function to reference images packaged within the WizSDK module
    
    Returns:
        Full file path to the packaged image.
    """
    return os.path.dirname(__file__) + "/images/" + filename
